// This file is part of rust-u4pak.
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at https://mozilla.org/MPL/2.0/.

use std::{fs::OpenOptions, io::{BufWriter, Read, Seek, SeekFrom, Write}, num::NonZeroUsize, path::{Path, PathBuf}};
use std::fs::File;
use std::{cmp, str};
use std::thread as std_thread;
use std::time::Duration as std_time_Duration;

//use File::sync_all;

use crossbeam_channel::{Receiver, Sender, unbounded};
use crossbeam_utils::thread;
use flate2::bufread::ZlibDecoder;
use aes::BLOCK_SIZE;

use crate::util::align;
use crate::decrypt::decrypt;

use crate::{Error, Result, Pak, pak::{self, COMPR_NONE, PAK_RELATIVE_COMPRESSION_OFFSET_VERSION, Variant, compression_method_name}, util::parse_pak_path};
use crate::Record;
use crate::Filter;
use crate::reopen::Reopen;
use log::{debug};

#[derive(Debug)]
pub struct UnpackOptions<'a> {
    pub dirname_from_compression: bool,
    pub verbose: bool,
    pub null_separated: bool,
    pub paths: Option<&'a [&'a str]>,
    pub thread_count: NonZeroUsize,
    pub encryption_key: Option<Vec<u8>>,
}

impl Default for UnpackOptions<'_> {
    fn default() -> Self {
        Self {
            dirname_from_compression: false,
            verbose: false,
            null_separated: false,
            paths: None,
            thread_count: NonZeroUsize::new(num_cpus::get()).unwrap_or(NonZeroUsize::new(1).unwrap()),
            encryption_key: None,
        }
    }
}

#[inline]
fn unpack_iter<'a>(pak: &Pak, in_file: &mut File, outdir: &Path, options: &'a UnpackOptions<'a>, records_iter: impl Iterator<Item=&'a Record>) -> Result<()> {
    let version = pak.version();
    let variant = pak.variant();

    let dirnames = if options.dirname_from_compression {
        let mut zlib_outdir = outdir.to_path_buf();
        let mut none_outdir = outdir.to_path_buf();

        zlib_outdir.push("zlib");
        none_outdir.push("none");

        Some((zlib_outdir, none_outdir))
    } else {
        None
    };

    let pak_path = in_file.path()?;

    let thread_result = thread::scope::<_, Result<()>>(|scope| {
        let (work_sender, work_receiver) = unbounded();
        let (result_sender, result_receiver) = unbounded();

        for _ in 0..options.thread_count.get() {
            let work_receiver = work_receiver.clone();
            let result_sender = result_sender.clone();
            let mut in_file = File::open(&pak_path)?;

            scope.spawn(move |_| {
                let in_file = &mut in_file;
                if let Err(error) = worker_proc(in_file, version, variant, options.encryption_key.clone(), work_receiver, result_sender) {
                    if !error.error_type().is_channel_disconnected() {
                        eprintln!("error in worker thread: {}", error);
                    }
                }
            });
        }

        drop(work_receiver);
        drop(result_sender);

        if let Some((zlib_outdir, none_outdir)) = &dirnames {
            for record in records_iter {
                let method = record.compression_method();
                let outdir = if method == COMPR_NONE { &none_outdir } else { &zlib_outdir };

                match work_sender.send(Work { record, outdir }) {
                    Ok(()) => {}
                    //Err(error) =>
                    //    return Err(Error::new(error.to_string()).with_path(record.filename()))
					Err(error) => {}
                }
            }
        } else {
            for record in records_iter {
                match work_sender.send(Work { record, outdir }) {
                    Ok(()) => {}
                    //Err(error) =>
                    //    return Err(Error::new(error.to_string()).with_path(record.filename()))
					Err(error) => {}
                }
            }
        }

        drop(work_sender);

        #[cfg(target_family="unix")]
        let mut stdout = std::io::stdout();

        let linesep = if options.null_separated { '\0' } else { '\n' };

        while let Ok(result) = result_receiver.recv() {
            let path = result?;
            if options.verbose {
                #[cfg(target_family="unix")]
                {
                    use std::os::unix::ffi::OsStrExt;
                    let _ = stdout.write_all(path.as_os_str().as_bytes());
                    let _ = stdout.write_all(&[linesep as u8]);
                }

                #[cfg(not(target_family="unix"))]
                {
                    print!("{}{}", path.to_string_lossy(), linesep);
                }
            }
        }

        drop(result_receiver);

        Ok(())
    });

    match thread_result {
        Err(error) => {
            return Err(Error::new(format!("threading error: {:?}", error)));
        }
        Ok(result) => result
    }
}

#[link(name = "libooz", kind = "static")]
extern "C" {
    fn Ooz_Decompress(src_buf: *const u8, src_len: u32, dst: *mut u8, dst_size: usize) -> i32;
}

pub fn decompress_oodle(source: *const u8, src_len: usize, destination: *mut u8, dst_size: usize) -> i32 {
    unsafe {
        return Ooz_Decompress(source, src_len as u32, destination, dst_size);
    }
}


pub fn unpack<'a>(pak: &Pak, in_file: &mut File, outdir: impl AsRef<Path>, options: UnpackOptions<'a>) -> Result<()> {
    let outdir = outdir.as_ref();

    if let Some(paths) = options.paths {
        let mut filter: Filter = paths.into();
        let records = pak.index().records().iter()
            .filter(|record| filter.visit(record.filename()));

        unpack_iter(pak, in_file, outdir, &options, records)?;
        filter.assert_all_visited()?;
    } else {
        unpack_iter(pak, in_file, outdir, &options, pak.index().records().iter())?;
    }
    Ok(())
}

pub fn get_oodle_compression_type(byte1: u8, byte2: u8, filename: String) -> String {
	let mut oodle_compression_type = String::new();
	oodle_compression_type = "None".to_string();
	if byte1 == 0x8c {
		if byte2 == 0x06 {
			oodle_compression_type = "Kraken".to_string();
		}
		if byte2 == 0x0a {
			oodle_compression_type = "Selkie or Mermaid".to_string();
		}
		if byte2 == 0x0c {
			oodle_compression_type = "Leviathan".to_string();
		}
		if oodle_compression_type == "None" {
			oodle_compression_type = "Unknown".to_string();
			println!("Warning: file {filename} appears to use an unrecognized type of Oodle compression ({byte2:#x})");
		}
		/*else {
			println!("File {filename} uses the {oodle_compression_type} variation of Oodle compression ({byte2:#x})");
		}*/
	}
	else {
		println!("Warning: file {filename} is marked as compressed using Oodle, but does not appear to be compressed and will be output as-is ({byte1:#x} {byte2:#x})");
	}
	return oodle_compression_type;
}

pub fn unpack_record(record: &Record, version: u32, variant: Variant, in_file: &mut File, outdir: impl AsRef<Path>, encryption_key: Option<Vec<u8>>) -> Result<PathBuf> {
    let header_size = pak::Pak::header_size(version, variant, record);
    
    let mut path = outdir.as_ref().to_path_buf();
    for component in parse_pak_path(record.filename()) {
        path.push(component);
    }
    
    let mut out_file = match OpenOptions::new()
            .write(true)
            .create(true)
            .truncate(true)
            .open(&path) {
        Ok(file) => file,
		Err(error) => {
            if error.kind() == std::io::ErrorKind::NotFound {
                if let Some(parent) = path.parent() {
                    std::fs::create_dir_all(parent)?;
                    OpenOptions::new().write(true).create(true).open(&path)?
                } else {
                    return Err(Error::io_with_path(error, path));
                }
            } else {
                return Err(Error::io_with_path(error, path));
            }
        }
    };

    let start_offset = record.offset() + header_size;
    in_file.seek(SeekFrom::Start(start_offset))?;

    // Encrypted files need to be read in 16 byte blocks
    let buffer_length = if record.encrypted() {
        align(record.size(), BLOCK_SIZE as u64)
    } else {
        record.size()
    } as usize;

    let mut in_buffer = vec![0u8; buffer_length];
    in_file.read_exact(&mut in_buffer)?;
    
    decrypt_entry(&mut in_buffer, record, encryption_key, record.size() as usize)?;
    debug!("unpacking {:?}", record);
    //println!("unpacking {:?}", record);
	let mut record_filename = record.filename();
	let mut compression_block_size = record.compression_block_size();
	
	let mut out_buffer = Vec::with_capacity(compression_block_size as usize);
	let mut oodle_src = vec![0u8; compression_block_size as usize];
	let mut chunk_dst = vec![0u8; compression_block_size as usize];
	
    match record.compression_method() {
        pak::COMPR_NONE => {
			//println!("Extracting uncompressed file: {record_filename}");
            out_file.write_all(&in_buffer)?;
            out_file.flush()?;
        }
		pak::COMPR_OODLE => {
			//println!("Decompressing Oodle file: {record_filename}");
			let record_uncompressed_size = record.uncompressed_size();
            if let Some(blocks) = record.compression_blocks() {
				let num_blocks = blocks.len();
				//println!("Oodle file {record_filename} has {num_blocks} block(s)");
                //let mut out_file = BufWriter::new(out_file);
				
				let mut use_decompressor = 1;
				let mut block_number = 1;
				let mut byte_number = 0;
				let mut decompress_result = -1;
				// the ooz library will sometimes write past the end of the buffer if the buffer is only big enough for the data
				// the number of extra bytes is probably really excessive, but the code will crash if there are not enough extra bytes
				let extra_buffer_space = 0x800000;
				let mut total_decompressed_byte_count = 0;
				
                for block in blocks {
					//println!("Decompressing block {block_number} of {num_blocks} for file {record_filename}");
                    let mut block_start = (block.start_offset - header_size) as usize;
                    let mut block_end = (block.end_offset - header_size) as usize;

					let mut dst_size = cmp::min(record_uncompressed_size, compression_block_size as u64) as usize;
					
					// libooz will return an error when trying to decompress the last block of a multi-block file unless the expected data size matches
					if block_number == num_blocks {
						let remaining_byte_count = (record_uncompressed_size - total_decompressed_byte_count) as usize;
						if remaining_byte_count > 0 {
							dst_size = cmp::min(dst_size, remaining_byte_count) as usize;
						}
						else {
							println!("Error: got zero or negative remaining byte count for block {block_number} of {num_blocks} for file {record_filename}");
						}
					}
				
                    if version < PAK_RELATIVE_COMPRESSION_OFFSET_VERSION {
                        block_start -= (start_offset - header_size) as usize;
                        block_end -= (start_offset - header_size) as usize;
                    }
					
                    out_buffer.clear();
					let src_size = (block_end - block_start) as usize;
					oodle_src = vec![0u8; src_size];
					
					let debug_start_offset = start_offset + block_start as u64;
					let debug_end_offset = start_offset + block_end as u64;
					
					let hex_data_1 = in_buffer[block_start];
					let hex_data_2 = in_buffer[block_start+1];
					
					//println!("Block {block_number} of {num_blocks} for file {record_filename} is from offset {debug_start_offset} to {debug_end_offset}, with length {src_size}, first two bytes are {hex_data_1:#x} {hex_data_2:#x}");
					
					byte_number = 0;
					while byte_number < src_size {
						//oodle_src.push(in_buffer[block_start + byte_number]);
						oodle_src[byte_number] = in_buffer[block_start + byte_number];
						byte_number += 1;
					}
					
					// Oodle appears to not compress very small files
					// I haven't been able to determine yet if it adds a "not compressed" header of any kind or just writes the raw data 
					//if block_number == 1 {
						let mut oodle_compression_type = get_oodle_compression_type(hex_data_1, hex_data_2, record_filename.to_string());
						
						if oodle_compression_type == "None" {
							use_decompressor = 0;
						}
					//}
					chunk_dst = vec![0u8; dst_size + extra_buffer_space];
					let mut len_out_buffer = out_buffer.len();
					let mut len_chunk_dst = dst_size;
					
					if use_decompressor == 1 {	
						decompress_result = decompress_oodle(
							oodle_src.as_ptr(),
							src_size,
							chunk_dst.as_mut_ptr(),
							dst_size,
						);						
						
						if decompress_result < 0 {
							println!("Error decompressing Oodle block - code: {decompress_result} into buffer of size {dst_size}, first two bytes are ({hex_data_1:#x} {hex_data_2:#x})");
						}
						else {
							//println!("Decompressed {decompress_result} bytes into buffer of size {dst_size} for file {record_filename}");
							total_decompressed_byte_count += decompress_result as u64;
						}
						byte_number = 0;
						len_out_buffer = out_buffer.len();
						//len_chunk_dst = chunk_dst.len();
						len_chunk_dst = dst_size;
						//println!("len(out_buffer): {len_out_buffer} len(chunk_dst): {len_chunk_dst} compression_block_size: {compression_block_size} for block {block_number} of file {record_filename}");
						while byte_number < dst_size {
							out_buffer.push(chunk_dst[byte_number]);
							byte_number += 1;
						}
						len_out_buffer = out_buffer.len();
						//println!("copied {byte_number} bytes - len(out_buffer): {len_out_buffer} len(chunk_dst): {len_chunk_dst} compression_block_size: {compression_block_size} for block {block_number} of file {record_filename}");
						
						//println!("Writing decompressed block {block_number} of file {record_filename} to disk");
						out_file.write_all(&out_buffer)?;
						out_file.flush()?;
						out_buffer.clear();
						//println!("Writing complete for decompressed block {block_number} of file {record_filename}");
					}
					else
					{
						//if block_number > 1 {
							//if block_number == num_blocks {
								println!("Warning: outputting raw data for block {block_number} of {num_blocks} for file {record_filename} as it does not appear to be compressed");
							//}
						//}
						out_file.write_all(&in_buffer)?;
						out_file.flush()?;
						//println!("Writing complete for raw data for file {record_filename}");
					}
					//println!("Finished decompressing block {block_number} of {num_blocks} for file {record_filename}");
					block_number += 1;
                }
				/*println!("Flushing file {record_filename}");
                out_file.flush()?;
				println!("Sleeping 10 seconds");
				std_thread::sleep(std_time_Duration::from_millis(10000));*/
				//println!("Wrote Oodle-decompressed file {record_filename} to disk");
            } else {
				println!("Error: this should never happen with Returnal");
            }
			//println!("Done handling Oodle file: {record_filename}");
        }
        pak::COMPR_ZLIB => {
			//println!("Decompressing zlib file: {record_filename}");
            if let Some(blocks) = record.compression_blocks() {
                let mut out_file = BufWriter::new(out_file);

                let mut out_buffer = Vec::with_capacity(record.compression_block_size() as usize);

                for block in blocks {
                    let mut block_start = (block.start_offset - header_size) as usize;
                    let mut block_end = (block.end_offset - header_size) as usize;

                    if version < PAK_RELATIVE_COMPRESSION_OFFSET_VERSION {
                        block_start -= (start_offset - header_size) as usize;
                        block_end -= (start_offset - header_size) as usize;
                    }

                    let mut zlib = ZlibDecoder::new(&in_buffer[block_start..block_end]);
                    out_buffer.clear();
                    zlib.read_to_end(&mut out_buffer)?;
                    out_file.write_all(&out_buffer)?;
                }
                out_file.flush()?;
				//println!("Wrote zlib-decompressed file {record_filename} to disk");
            } else {
                // version 2 has compression support, but not compression blocks
                let mut out_buffer = Vec::new();

                let mut zlib = ZlibDecoder::new(&in_buffer[..]);
                zlib.read_to_end(&mut out_buffer)?;
                out_file.write_all(&out_buffer)?;
                out_file.flush()?;
            }
        }
        _ => {
            return Err(Error::new(format!(
                    "unsupported compression method: {}",
                    compression_method_name(record.compression_method())))
                .with_path(record.filename()));
        }
    }
	//println!("Finished unpacking {:?}", record);
    Ok(path)
}

#[derive(Debug)]
struct Work<'a> {
    record: &'a Record,
    outdir: &'a Path,
}

fn worker_proc(in_file: &mut File, version: u32, variant: Variant, encryption_key: Option<Vec<u8>>, work_channel: Receiver<Work>, result_channel: Sender<Result<PathBuf>>) -> Result<()> {
    while let Ok(Work { record, outdir }) = work_channel.recv() {
        let result = unpack_record(record, version, variant, in_file, outdir, encryption_key.clone())
            .map_err(|error| error
                .with_path_if_none(record.filename()));

        result_channel.send(result)?;
    }

    Ok(())
}

fn decrypt_entry(buffer: &mut Vec<u8>, record: &Record, encryption_key: Option<Vec<u8>>, size: usize) -> Result<()> {
    if record.encrypted() {
        if let Some(key) = encryption_key {
            decrypt(buffer, &key);
            // Trim padded bytes from undersized encryption blocks
            buffer.truncate(size);
        } else {
            return Err(Error::new(
                "File is encrypted, but no encryption key was provided".to_string(),
            ).with_path(record.filename()));
        }
    }
    Ok(())
}